import asyncio
import time


class ExecutionTimer:
    """
    Context manager that enforces a minimum period between entries by sleeping.

    This is not a rate limiter like Throttler. With align_sleep=True, it aligns
    to wall-clock boundaries (e.g. each minute).

    Example usage:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_execution_timer.py
    """

    __slots__ = (
        "_period",
        "_align_sleep",
        "_start_time",
        "_next_time",
        "_start_wall",
    )

    def __init__(self, period: float = 60.0, align_sleep: bool = False):
        if not (isinstance(period, (int, float)) and period > 0.0):
            raise ValueError("`period` should be positive float")

        self._period = float(period)
        self._align_sleep = align_sleep

        self._start_time = 0.0
        self._next_time = 0.0
        self._start_wall = 0.0

    def _start(self):
        curr_time = time.monotonic()
        diff = self._next_time - curr_time
        return diff

    def _exit(self):
        if self._align_sleep:
            next_wall = self._start_wall + self._period
            next_wall -= self._start_wall % self._period
            wall_now = time.time()
            delay = max(0.0, next_wall - wall_now)
            self._next_time = time.monotonic() + delay
        else:
            self._next_time = self._start_time + self._period

    def __enter__(self):
        diff = self._start()
        if diff > 0.0:
            time.sleep(diff)
        self._start_time = time.monotonic()
        self._start_wall = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._exit()

    async def __aenter__(self):
        diff = self._start()
        if diff > 0.0:
            await asyncio.sleep(diff)
        self._start_time = time.monotonic()
        self._start_wall = time.time()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._exit()



import asyncio
import time

import aiohttp

from throttler import Throttler, ThrottlerSimultaneous


class SomeAPI:
    api_url = "https://example.com"

    def __init__(self, throttler):
        self.throttler = throttler

    async def request(self, session: aiohttp.ClientSession):
        async with self.throttler:
            async with session.get(self.api_url) as resp:
                return resp

    async def many_requests(self, count: int):
        async with aiohttp.ClientSession() as session:
            coros = [self.request(session) for _ in range(count)]
            for coro in asyncio.as_completed(coros):
                response = await coro
                print(f"{int(time.time())} | Result: {response.status}")


def example():
    async def run():
        api = SomeAPI(Throttler(rate_limit=3, period=1.0))
        await api.many_requests(100)

    asyncio.run(run())


def example_simultaneous():
    async def run():
        api = SomeAPI(ThrottlerSimultaneous(count=5))
        await api.many_requests(100)

    asyncio.run(run())



import asyncio


class ThrottlerSimultaneous:
    """
    Async context manager that limits concurrent access to a block.

    Example usages:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers.py
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers_aiohttp.py
    """

    __slots__ = ("_count", "_semaphore")

    def __init__(self, count: int):
        if not (isinstance(count, int) and count > 0):
            raise ValueError("`count` should be positive integer")

        self._count = count
        self._semaphore = None

    async def __aenter__(self):
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self._count)
        await self._semaphore.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self._semaphore.release()



import asyncio
import time

from throttler import Throttler, ThrottlerSimultaneous


async def task(throttler):
    async with throttler:
        return await asyncio.sleep(0.1)


async def many_tasks(throttler, count: int):
    coros = [task(throttler) for _ in range(count)]
    for coro in asyncio.as_completed(coros):
        _result = await coro
        print(f"Timestamp: {time.time()}")


def example():
    async def run():
        t = Throttler(rate_limit=10, period=3.0)
        await many_tasks(t, 100)

    asyncio.run(run())


def example_simultaneous():
    async def run():
        t = ThrottlerSimultaneous(count=5)
        await many_tasks(t, 100)

    asyncio.run(run())



import asyncio
import time
from collections import deque
from typing import Union


class Throttler:
    """
    Async context manager that limits how often the block can be entered.

    Example usages:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers.py
        - https://github.com/uburuntu/throttler/blob/master/examples/example_throttlers_aiohttp.py
    """

    __slots__ = (
        "_rate_limit",
        "_period",
        "_times",
    )

    def __init__(self, rate_limit: int, period: Union[int, float] = 1.0):
        if not (isinstance(rate_limit, int) and rate_limit > 0):
            raise ValueError("`rate_limit` should be positive integer")

        if not (isinstance(period, (int, float)) and period > 0.0):
            raise ValueError("`period` should be positive float")

        self._rate_limit = float(rate_limit)
        self._period = float(period)

        self._times = deque(0.0 for _ in range(rate_limit))

    async def __aenter__(self):
        while True:
            curr_ts = time.monotonic()
            diff = curr_ts - (self._times[0] + self._period)
            if diff > 0.0:
                self._times.popleft()
                break
            await asyncio.sleep(-diff)

        self._times.append(curr_ts)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



"""Zero-dependency Python package for easy throttling with asyncio support"""

from .execution_timer import ExecutionTimer
from .throttler import Throttler
from .throttler_simultaneous import ThrottlerSimultaneous
from .timer import Timer

from .decorators import (
    execution_timer,
    execution_timer_async,
    throttle,
    throttle_simultaneous,
    timer,
    timer_async,
)

__author__ = "uburuntu"
__email__ = "github@rmbk.me"

__license__ = "MIT"
__version__ = "1.2.3"



import asyncio

from throttler import (
    ExecutionTimer,
    Throttler,
    ThrottlerSimultaneous,
    Timer,
    execution_timer,
    execution_timer_async,
    throttle,
    throttle_simultaneous,
    timer,
    timer_async,
)


def sync_smoke():
    with ExecutionTimer(0.01):
        with Timer("smoke"):
            pass

    @execution_timer(0.01)
    def run():
        return "ok"

    @timer("smoke")
    def run_timer():
        return "ok"

    assert run() == "ok"
    assert run_timer() == "ok"


async def async_smoke():
    throttler = Throttler(1, 0.01)
    async with throttler:
        pass

    throttler_simultaneous = ThrottlerSimultaneous(1)
    async with throttler_simultaneous:
        pass

    @throttle(1, 0.01)
    async def run():
        return "ok"

    @throttle_simultaneous(1)
    async def run_simultaneous():
        return "ok"

    @execution_timer_async(0.01)
    async def run_execution_timer():
        return "ok"

    @timer_async("smoke")
    async def run_timer():
        return "ok"

    assert await run() == "ok"
    assert await run_simultaneous() == "ok"
    assert await run_execution_timer() == "ok"
    assert await run_timer() == "ok"


def main():
    sync_smoke()
    asyncio.run(async_smoke())


if __name__ == "__main__":
    main()



import os
import time
from math import isclose

import pytest

from throttler import ExecutionTimer, execution_timer, execution_timer_async

# Weak machines may be used for CI, causing delays
ABS_TOL = 0.2 if os.getenv("CI") else 0.1


class TestExecutionTimer:
    @pytest.mark.parametrize("period", (-1, 0, 0.0, "bad"))
    def test_invalid_period(self, period):
        with pytest.raises(ValueError):
            ExecutionTimer(period)

    @pytest.mark.parametrize(("period",), ((1.0,), (3.0,)))
    def test_without_align(self, period: float):
        @execution_timer(period)
        def t():
            curr_ts = time.time()
            if i > 0:
                assert isclose(curr_ts - prev_ts, period, abs_tol=ABS_TOL)
            return curr_ts

        prev_ts = None
        for i in range(3):
            prev_ts = t()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(("period",), ((1.0,), (3.0,)))
    async def test_without_align_via_async_decorator(self, period: float):
        @execution_timer(period)
        async def t():
            curr_ts = time.time()
            if i > 0:
                assert isclose(curr_ts - prev_ts, period, abs_tol=ABS_TOL)
            return curr_ts

        prev_ts = None
        for i in range(3):
            prev_ts = await t()

    @pytest.mark.parametrize(("period",), ((3.0,), (5.0,)))
    def test_with_align(self, period: float):
        @execution_timer(period, align_sleep=True)
        def t():
            curr_ts = time.time()
            if i > 0:
                assert isclose(curr_ts % period, 0.0, abs_tol=ABS_TOL)
                if i > 1:
                    assert isclose(curr_ts - prev_ts, period, abs_tol=ABS_TOL)
            return curr_ts

        prev_ts = None
        for i in range(3):
            prev_ts = t()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(("period",), ((1.0,), (3.0,)))
    async def test_without_align_async(self, period: float):
        @execution_timer_async(period)
        async def t():
            curr_ts = time.time()
            if i > 0:
                assert isclose(curr_ts - prev_ts, period, abs_tol=ABS_TOL)
            return curr_ts

        prev_ts = None
        for i in range(3):
            prev_ts = await t()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(("period",), ((3.0,), (5.0,)))
    async def test_with_align_async(self, period: float):
        @execution_timer_async(period, align_sleep=True)
        async def t():
            curr_ts = time.time()
            if i > 0:
                assert isclose(curr_ts % period, 0.0, abs_tol=ABS_TOL)
                if i > 1:
                    assert isclose(curr_ts - prev_ts, period, abs_tol=ABS_TOL)
            return curr_ts

        prev_ts = None
        for i in range(3):
            prev_ts = await t()



from datetime import datetime
import time
from typing import Callable


class Timer:
    """
    Context manager for printing start, end, elapsed and average times.

    Elapsed timing uses a monotonic clock, while start/end timestamps are wall time.

    Example usage:
        - https://github.com/uburuntu/throttler/blob/master/examples/example_timer.py
    """

    def __init__(
        self, name: str = None, verbose: bool = False, print_func: Callable = None
    ):
        self.iteration = 1
        self.start_dt = None
        self.start_perf = 0.0
        self.elapsed_all = 0.0

        self.name = name
        self.verbose = verbose
        self.print = print_func or print

    def __enter__(self):
        self.start_dt = datetime.now()
        self.start_perf = time.perf_counter()
        if self.verbose:
            self.print(
                f"{f'#{self.iteration}':>5} | {self.name or 'Timer'} | begin: {self.start_dt}"
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        curr_dt = datetime.now()
        elapsed = time.perf_counter() - self.start_perf

        self.elapsed_all += elapsed
        average = self.elapsed_all / self.iteration

        if self.verbose:
            self.print(
                f"{f'#{self.iteration}':>5} | {self.name or 'Timer'} |   end: {curr_dt}, elapsed: {elapsed:.2f} sec, "
                f"average: {average:.2f} sec\n"
            )
        else:
            self.print(f"{self.name or 'Timer'} | elapsed: {elapsed:.2f} sec")

        self.iteration += 1



import asyncio

import pytest

from tests.service import ServiceSimultaneous
from throttler import ThrottlerSimultaneous, throttle_simultaneous


class TestThrottlerSimultaneous:
    @pytest.mark.parametrize("count", (0, -1, -5))
    def test_invalid_count(self, count: int):
        with pytest.raises(ValueError):
            ThrottlerSimultaneous(count)

    @pytest.mark.parametrize(
        ("max_simultaneous", "count"), ((1, 10), (3, 10), (100, 500))
    )
    def test_via_service_simultaneous(self, max_simultaneous: int, count: int):
        s = ServiceSimultaneous(max_simultaneous)

        @throttle_simultaneous(max_simultaneous)
        async def request(value: float):
            return await s.get(value)

        async def run():
            await asyncio.gather(*[request(v) for v in range(count)])

        asyncio.run(run())



import asyncio
import random
import time
from collections import deque


class BanException(Exception):
    pass


async def sleeper(value: float):
    s = random.uniform(0, 0.1)
    await asyncio.sleep(s)
    return value


class Service:
    def __init__(self, rate_limit: int, period: float):
        self.rate_limit = rate_limit
        self.period = period - 0.01

        self.times = deque((0.0,) * rate_limit)

    async def get(self, value: float):
        curr_ts = time.monotonic()
        diff = curr_ts - self.times[0]
        if diff < self.period:
            raise BanException("Limit exceeded")
        self.times.popleft()
        self.times.append(curr_ts)
        return await sleeper(value)


class ServiceSimultaneous:
    def __init__(self, max_simultaneous: int):
        self.max_simultaneous = max_simultaneous

        self.counter = 0

    async def get(self, value: float):
        self.counter += 1
        try:
            if self.counter > self.max_simultaneous:
                raise BanException("Limit exceeded")
            return await sleeper(value)
        finally:
            self.counter -= 1



import asyncio

import pytest

from tests.service import BanException, Service, ServiceSimultaneous


class TestService:
    @pytest.mark.parametrize(("rate_limit", "period"), ((1, 0.5), (3, 1.0), (100, 1.5)))
    def test_service(self, rate_limit: int, period: float):
        s = Service(rate_limit, period)

        async def request(value: float):
            return await s.get(value)

        async def run():
            await asyncio.gather(
                *[request(v) for v in range(int(rate_limit / period) + 100)]
            )

        with pytest.raises(BanException):
            asyncio.run(run())

    @pytest.mark.parametrize(("max_simultaneous",), ((1,), (3,), (100,)))
    def test_service_simultaneous(self, max_simultaneous: int):
        s = ServiceSimultaneous(max_simultaneous)

        async def request(value: float):
            return await s.get(value)

        async def run():
            await asyncio.gather(*[request(v) for v in range(max_simultaneous + 100)])

        with pytest.raises(BanException):
            asyncio.run(run())



import asyncio
from itertools import product

import pytest

from tests.service import Service
from throttler import throttle, Throttler


class TestThrottler:
    @pytest.mark.parametrize(
        ("rate_limit", "period"), tuple(product((-1, 0, "", 1), (-1, -1.0, 0, 0.0, "")))
    )
    def test_exceptions(self, rate_limit: int, period: float):
        with pytest.raises(ValueError):
            Throttler(rate_limit, period)

    @pytest.mark.parametrize(
        ("rate_limit", "period", "count"),
        tuple(product((1, 3, 5), (0.5, 1.0, 1.5), (3, 5, 7)))
        + tuple(product((100, 1000), (0.5, 1.0, 1.5), (10, 1000))),
    )
    def test_via_service(self, rate_limit: int, period: float, count: int):
        s = Service(rate_limit, period)

        @throttle(rate_limit, period)
        async def request(value: float):
            return await s.get(value)

        async def run():
            await asyncio.gather(*[request(v) for v in range(count)])

        asyncio.run(run())



import random
import time

from throttler import Timer


def example():
    """
    Output will be like:
        My Timer | elapsed: 0.85 sec
        My Timer | elapsed: 0.93 sec
        My Timer | elapsed: 0.31 sec
    """

    timer = Timer("My Timer")

    for _ in range(3):
        with timer:
            time.sleep(random.random())


def example_verbose():
    """
    Output will be like:
        #1 | My Timer | begin: 2020-03-26 01:46:07.648661
        #1 | My Timer |   end: 2020-03-26 01:46:08.382135, elapsed: 0.73 sec, average: 0.73 sec
        #2 | My Timer | begin: 2020-03-26 01:46:08.382135
        #2 | My Timer |   end: 2020-03-26 01:46:08.599919, elapsed: 0.22 sec, average: 0.48 sec
        #3 | My Timer | begin: 2020-03-26 01:46:08.599919
        #3 | My Timer |   end: 2020-03-26 01:46:09.083370, elapsed: 0.48 sec, average: 0.48 sec
    """

    timer = Timer("My Timer", verbose=True)

    for _ in range(3):
        with timer:
            time.sleep(random.random())



import time

import pytest

from throttler import timer, timer_async


class TestTimer:
    @pytest.mark.parametrize(("verbose",), ((True,), (False,)))
    def test_simple(self, verbose: bool):
        @timer(name="TestTimer", verbose=verbose)
        def t():
            print(time.time())

        for _ in range(3):
            t()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(("verbose",), ((True,), (False,)))
    async def test_simple_async(self, verbose: bool):
        @timer_async(name="TestTimer", verbose=verbose)
        async def t():
            print(time.time())

        for _ in range(3):
            await t()

    @pytest.mark.asyncio
    @pytest.mark.parametrize(("verbose",), ((True,), (False,)))
    async def test_timer_decorator_with_async_callable(self, verbose: bool):
        @timer(name="TestTimer", verbose=verbose)
        async def t():
            print(time.time())

        for _ in range(3):
            await t()



from functools import wraps
from typing import Callable
import inspect

from .execution_timer import ExecutionTimer
from .throttler import Throttler
from .throttler_simultaneous import ThrottlerSimultaneous
from .timer import Timer


def throttle(rate_limit: int, period: float = 1.0):
    """
    Decorator for limiting how often an async function can be entered.

    Args:
        rate_limit: Maximum number of calls per period.
        period: Time window in seconds.
    """

    def decorator(func):
        throttler = Throttler(rate_limit, period)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with throttler:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def throttle_simultaneous(count: int):
    """
    Decorator for limiting concurrent access to an async function.

    Args:
        count: Maximum number of simultaneous calls.
    """

    def decorator(func):
        throttler = ThrottlerSimultaneous(count)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with throttler:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def execution_timer(period: float = 60.0, align_sleep: bool = False):
    """
    Decorator that enforces a minimum period between function entries.

    This is not a rate limiter like Throttler. With align_sleep=True, it aligns
    to wall-clock boundaries (e.g. each minute).

    Supports both sync and async callables.
    """

    def decorator(func):
        et = ExecutionTimer(period, align_sleep)

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def wrapper(*args, **kwargs):
                async with et:
                    return await func(*args, **kwargs)

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                with et:
                    return func(*args, **kwargs)

        return wrapper

    return decorator


def execution_timer_async(period: float = 60.0, align_sleep: bool = False):
    """
    Async decorator that enforces a minimum period between function entries.

    This is not a rate limiter like Throttler. With align_sleep=True, it aligns
    to wall-clock boundaries (e.g. each minute).

    Backwards compatibility note: execution_timer() now supports async callables.
    """

    def decorator(func):
        et = ExecutionTimer(period, align_sleep)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            async with et:
                return await func(*args, **kwargs)

        return wrapper

    return decorator


def timer(name: str = None, verbose: bool = False, print_func: Callable = None):
    """
    Decorator for printing elapsed timing information around a function.

    Elapsed timing uses a monotonic clock, while start/end timestamps are wall time.

    Supports both sync and async callables.
    """

    def decorator(func):
        t = Timer(name, verbose, print_func)

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def wrapper(*args, **kwargs):
                with t:
                    return await func(*args, **kwargs)

        else:

            @wraps(func)
            def wrapper(*args, **kwargs):
                with t:
                    return func(*args, **kwargs)

        return wrapper

    return decorator


def timer_async(name: str = None, verbose: bool = False, print_func: Callable = None):
    """
    Async decorator for printing elapsed timing information around a function.

    Elapsed timing uses a monotonic clock, while start/end timestamps are wall time.

    Backwards compatibility note: timer() now supports async callables.
    """

    def decorator(func):
        t = Timer(name, verbose, print_func)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            with t:
                return await func(*args, **kwargs)

        return wrapper

    return decorator
